@@ -16,7 +16,8 @@ import {ViewUpdate, styleModule,
         exceptionSink, updateListener, logException,
         viewPlugin, ViewPlugin, PluginValue, PluginInstance, decorations, outerDecorations, atomicRanges,
         scrollMargins, MeasureRequest, editable, inputHandler, focusChangeEffect, perLineTextDirection,
-        scrollIntoView, UpdateFlag, ScrollTarget, bidiIsolatedRanges, getIsolatedRanges, scrollHandler} from "./extension"
+        scrollIntoView, UpdateFlag, ScrollTarget, bidiIsolatedRanges, getIsolatedRanges, scrollHandler,
+        clipboardInputFilter, clipboardOutputFilter} from "./extension"
 import {theme, darkTheme, buildTheme, baseThemeID, baseLightID, baseDarkID, lightDarkIDs, baseTheme} from "./theme"
 import {DOMObserver} from "./domobserver"
 import {Attrs, updateAttrs, combineAttrs} from "./attributes"
@@ -214,6 +215,7 @@ export class EditorView {
     this.updateState = UpdateState.Idle
 
     this.requestMeasure()
+    if (document.fonts?.ready) document.fonts.ready.then(() => this.requestMeasure())
   }
 
   /// All regular editor state updates should go through this. It
@@ -496,7 +498,8 @@ export class EditorView {
       this.measureScheduled = -1
     }
 
-    if (updated && !updated.empty) for (let listener of this.state.facet(updateListener)) listener(updated)
+    if (updated && !updated.empty)
+      for (let listener of this.state.facet(updateListener)) listener(updated)
   }
 
   /// Get the CSS classes for the currently active editor themes.
@@ -635,7 +638,7 @@ export class EditorView {
 
   /// Find the line block around the given document position. A line
   /// block is a range delimited on both sides by either a
-  /// non-[hidden](#view.Decoration^replace) line breaks, or the
+  /// non-[hidden](#view.Decoration^replace) line break, or the
   /// start/end of the document. It will usually just hold a line of
   /// text, but may be broken into multiple textblocks by block
   /// widgets.
@@ -849,6 +852,7 @@ export class EditorView {
   /// plugins. The view instance can no longer be used after
   /// calling this.
   destroy() {
+    if (this.root.activeElement == this.contentDOM) this.contentDOM.blur()
     for (let plugin of this.plugins) plugin.destroy(this)
     this.plugins = []
     this.inputState.destroy()
@@ -903,6 +907,24 @@ export class EditorView {
                                               ref.top - scrollTop, scrollLeft, true))
   }
 
+  /// Enable or disable tab-focus mode, which disables key bindings
+  /// for Tab and Shift-Tab, letting the browser's default
+  /// focus-changing behavior go through instead. This is useful to
+  /// prevent trapping keyboard users in your editor.
+  ///
+  /// Without argument, this toggles the mode. With a boolean, it
+  /// enables (true) or disables it (false). Given a number, it
+  /// temporarily enables the mode until that number of milliseconds
+  /// have passed or another non-Tab key is pressed.
+  setTabFocusMode(to?: boolean | number) {
+    if (to == null)
+      this.inputState.tabFocusMode = this.inputState.tabFocusMode < 0 ? 0 : -1
+    else if (typeof to == "boolean")
+      this.inputState.tabFocusMode = to ? 0 : -1
+    else if (this.inputState.tabFocusMode != 0)
+      this.inputState.tabFocusMode = Date.now() + to
+  }
+
   /// Facet to add a [style
   /// module](https://github.com/marijnh/style-mod#documentation) to
   /// an editor view. The view will ensure that the module is
@@ -945,6 +967,13 @@ export class EditorView {
   /// dispatching the custom behavior as a separate transaction.
   static inputHandler = inputHandler
 
+  /// Functions provided in this facet will be used to transform text
+  /// pasted or dropped into the editor.
+  static clipboardInputFilter = clipboardInputFilter
+
+  /// Transform text copied or dragged from the editor.
+  static clipboardOutputFilter = clipboardOutputFilter
+
   /// Scroll handlers can override how things are scrolled into view.
   /// If they return `true`, no further handling happens for the
   /// scrolling. If they return false, the default scroll behavior is
