@@ -30,6 +30,9 @@ export const inputHandler = Facet.define<(view: EditorView, from: number, to: nu
 
 export const focusChangeEffect = Facet.define<(state: EditorState, focusing: boolean) => StateEffect<any> | null>()
 
+export const clipboardInputFilter = Facet.define<(text: string, state: EditorState) => string>()
+export const clipboardOutputFilter = Facet.define<(text: string, state: EditorState) => string>()
+
 export const perLineTextDirection = Facet.define<boolean, boolean>({
   combine: values => values.some(x => x)
 })
@@ -73,6 +76,8 @@ export class ScrollTarget {
 
 export const scrollIntoView = StateEffect.define<ScrollTarget>({map: (t, ch) => t.map(ch)})
 
+export const setEditContextFormatting = StateEffect.define<DecorationSet>()
+
 /// Log or report an unhandled exception in client code. Should
 /// probably only be used by extension code that allows client code to
 /// provide functions, and calls those functions in a context where an
