@@ -43,6 +43,11 @@ GutterMarker.prototype.point = true
 /// in all gutters for the line).
 export const gutterLineClass = Facet.define<RangeSet<GutterMarker>>()
 
+/// Facet used to add a class to all gutter elements next to a widget.
+/// Should not provide widgets with a `toDOM` method.
+export const gutterWidgetClass =
+  Facet.define<(view: EditorView, widget: WidgetType, block: BlockInfo) => GutterMarker | null>()
+
 type Handlers = {[event: string]: (view: EditorView, line: BlockInfo, event: Event) => boolean}
 
 interface GutterConfig {
@@ -272,8 +277,12 @@ class UpdateContext {
   }
 
   widget(view: EditorView, block: BlockInfo) {
-    let marker = this.gutter.config.widgetMarker(view, block.widget!, block)
-    if (marker) this.addElement(view, block, [marker])
+    let marker = this.gutter.config.widgetMarker(view, block.widget!, block), markers = marker ? [marker] : null
+    for (let cls of view.state.facet(gutterWidgetClass)) {
+      let marker = cls(view, block.widget!, block)
+      if (marker) (markers || (markers = [])).push(marker)
+    }
+    if (markers) this.addElement(view, block, markers)
   }
 
   finish() {
@@ -410,6 +419,9 @@ interface LineNumberConfig {
 /// Facet used to provide markers to the line number gutter.
 export const lineNumberMarkers = Facet.define<RangeSet<GutterMarker>>()
 
+/// Facet used to create markers in the line number gutter next to widgets.
+export const lineNumberWidgetMarker = Facet.define<(view: EditorView, widget: WidgetType, block: BlockInfo) => GutterMarker | null>()
+
 const lineNumberConfig = Facet.define<LineNumberConfig, Required<LineNumberConfig>>({
   combine(values) {
     return combineConfig<Required<LineNumberConfig>>(values, {formatNumber: String, domEventHandlers: {}}, {
@@ -445,7 +457,13 @@ const lineNumberGutter = activeGutters.compute([lineNumberConfig], state => ({
     if (others.some(m => m.toDOM)) return null
     return new NumberMarker(formatNumber(view, view.state.doc.lineAt(line.from).number))
   },
-  widgetMarker: () => null,
+  widgetMarker: (view, widget, block) => {
+    for (let m of view.state.facet(lineNumberWidgetMarker)) {
+      let result = m(view, widget, block)
+      if (result) return result
+    }
+    return null
+  },
   lineMarkerChange: update => update.startState.facet(lineNumberConfig) != update.state.facet(lineNumberConfig),
   initialSpacer(view: EditorView) {
     return new NumberMarker(formatNumber(view, maxLineNumber(view.state.doc.lines)))
