@@ -1,12 +1,14 @@
-import {EditorSelection, EditorState, SelectionRange, RangeSet, Annotation, Text} from "@codemirror/state"
+import {EditorSelection, EditorState, SelectionRange, RangeSet, Annotation, Text, Facet} from "@codemirror/state"
 import {EditorView} from "./editorview"
 import {ContentView} from "./contentview"
 import {LineView} from "./blockview"
 import {ViewUpdate, PluginValue, clickAddsSelectionRange, dragMovesSelection as dragBehavior, atomicRanges,
-        logException, mouseSelectionStyle, PluginInstance, focusChangeEffect, getScrollMargins} from "./extension"
+        logException, mouseSelectionStyle, PluginInstance, focusChangeEffect, getScrollMargins,
+        clipboardInputFilter, clipboardOutputFilter} from "./extension"
 import browser from "./browser"
 import {groupAt, skipAtomicRanges} from "./cursor"
-import {getSelection, focusPreventScroll, Rect, dispatchKey, scrollableParent} from "./dom"
+import {getSelection, focusPreventScroll, Rect, dispatchKey, scrollableParents} from "./dom"
+import {applyDOMChangeInner} from "./domchange"
 
 // This will also be where dragging info and such goes
 export class InputState {
@@ -22,9 +24,15 @@ export class InputState {
   // avoid messing up the virtual keyboard state.
   pendingIOSKey: undefined | {key: string, keyCode: number} | KeyboardEvent = undefined
 
+  /// When enabled (>-1), tab presses are not given to key handlers,
+  /// leaving the browser's default behavior. If >0, the mode expires
+  /// at that timestamp, and any other keypress clears it.
+  /// Esc enables temporary tab focus mode for two seconds when not
+  /// otherwise handled.
+  tabFocusMode: number = -1
+
   lastSelectionOrigin: string | null = null
   lastSelectionTime: number = 0
-  lastEscPress: number = 0
   lastContextMenu: number = 0
   scrollHandlers: ((event: Event) => boolean | void)[] = []
 
@@ -112,8 +120,10 @@ export class InputState {
     this.lastKeyCode = event.keyCode
     this.lastKeyTime = Date.now()
 
-    if (event.keyCode == 9 && Date.now() < this.lastEscPress + 2000) return true
-    if (event.keyCode != 27 && modifierCodes.indexOf(event.keyCode) < 0) this.view.inputState.lastEscPress = 0
+    if (event.keyCode == 9 && this.tabFocusMode > -1 && (!this.tabFocusMode || Date.now() <= this.tabFocusMode))
+      return true
+    if (this.tabFocusMode > 0 && event.keyCode != 27 && modifierCodes.indexOf(event.keyCode) < 0)
+      this.tabFocusMode = -1
 
     // Chrome for Android usually doesn't fire proper key events, but
     // occasionally does, usually surrounded by a bunch of complicated
@@ -173,6 +183,7 @@ export class InputState {
   }
 
   update(update: ViewUpdate) {
+    this.view.observer.update(update)
     if (this.mouseSelection) this.mouseSelection.update(update)
     if (this.draggedContent && update.docChanged) this.draggedContent = this.draggedContent.map(update.changes)
     if (update.transactions.length) this.lastKeyCode = this.lastSelectionTime = 0
@@ -278,7 +289,7 @@ class MouseSelection {
   extend: boolean
   multiple: boolean
   lastEvent: MouseEvent
-  scrollParent: HTMLElement | null
+  scrollParents: {x?: HTMLElement, y?: HTMLElement}
   scrollSpeed = {x: 0, y: 0}
   scrolling = -1
   atoms: readonly RangeSet<any>[]
@@ -288,7 +299,7 @@ class MouseSelection {
               private style: MouseSelectionStyle,
               private mustSelect: boolean) {
     this.lastEvent = startEvent
-    this.scrollParent = scrollableParent(view.contentDOM)
+    this.scrollParents = scrollableParents(view.contentDOM)
     this.atoms = view.state.facet(atomicRanges).map(f => f(view))
     let doc = view.contentDOM.ownerDocument!
     doc.addEventListener("mousemove", this.move = this.move.bind(this))
@@ -311,18 +322,19 @@ class MouseSelection {
     this.select(this.lastEvent = event)
 
     let sx = 0, sy = 0
-    let rect = this.scrollParent?.getBoundingClientRect()
-      || {left: 0, top: 0, right: this.view.win.innerWidth, bottom: this.view.win.innerHeight}
+    let left = 0, top = 0, right = this.view.win.innerWidth, bottom = this.view.win.innerHeight
+    if (this.scrollParents.x) ({left, right} = this.scrollParents.x.getBoundingClientRect())
+    if (this.scrollParents.y) ({top, bottom} = this.scrollParents.y.getBoundingClientRect())
     let margins = getScrollMargins(this.view)
 
-    if (event.clientX - margins.left <= rect.left + dragScrollMargin)
-      sx = -dragScrollSpeed(rect.left - event.clientX)
-    else if (event.clientX + margins.right >= rect.right - dragScrollMargin)
-      sx = dragScrollSpeed(event.clientX - rect.right)
-    if (event.clientY - margins.top <= rect.top + dragScrollMargin)
-      sy = -dragScrollSpeed(rect.top - event.clientY)
-    else if (event.clientY + margins.bottom >= rect.bottom - dragScrollMargin)
-      sy = dragScrollSpeed(event.clientY - rect.bottom)
+    if (event.clientX - margins.left <= left + dragScrollMargin)
+      sx = -dragScrollSpeed(left - event.clientX)
+    else if (event.clientX + margins.right >= right - dragScrollMargin)
+      sx = dragScrollSpeed(event.clientX - right)
+    if (event.clientY - margins.top <= top + dragScrollMargin)
+      sy = -dragScrollSpeed(top - event.clientY)
+    else if (event.clientY + margins.bottom >= bottom - dragScrollMargin)
+      sy = dragScrollSpeed(event.clientY - bottom)
     this.setScrollSpeed(sx, sy)
   }
 
@@ -351,12 +363,16 @@ class MouseSelection {
   }
 
   scroll() {
-    if (this.scrollParent) {
-      this.scrollParent.scrollLeft += this.scrollSpeed.x
-      this.scrollParent.scrollTop += this.scrollSpeed.y
-    } else {
-      this.view.win.scrollBy(this.scrollSpeed.x, this.scrollSpeed.y)
+    let {x, y} = this.scrollSpeed
+    if (x && this.scrollParents.x) {
+      this.scrollParents.x.scrollLeft += x
+      x = 0
     }
+    if (y && this.scrollParents.y) {
+      this.scrollParents.y.scrollTop += y
+      y = 0
+    }
+    if (x || y) this.view.win.scrollBy(x, y)
     if (this.dragging === false) this.select(this.lastEvent)
   }
 
@@ -392,7 +408,10 @@ class MouseSelection {
   }
 
   update(update: ViewUpdate) {
-    if (this.style.update(update)) setTimeout(() => this.select(this.lastEvent), 20)
+    if (update.transactions.some(tr => tr.isUserEvent("input.type")))
+      this.destroy()
+    else if (this.style.update(update))
+      setTimeout(() => this.select(this.lastEvent), 20)
   }
 }
 
@@ -453,7 +472,13 @@ function capturePaste(view: EditorView) {
   }, 50)
 }
 
+function textFilter(state: EditorState, facet: Facet<(value: string, state: EditorState) => string>, text: string) {
+  for (let filter of state.facet(facet)) text = filter(text, state)
+  return text
+}
+
 function doPaste(view: EditorView, input: string) {
+  input = textFilter(view.state, clipboardInputFilter, input)
   let {state} = view, changes, i = 1, text = state.toText(input)
   let byLine = text.lines == state.selection.ranges.length
   let linewise = lastLinewiseCopy != null && state.selection.ranges.every(r => r.empty) && lastLinewiseCopy == text.toString()
@@ -489,7 +514,7 @@ observers.scroll = view => {
 
 handlers.keydown = (view, event: KeyboardEvent) => {
   view.inputState.setSelectionOrigin("select")
-  if (event.keyCode == 27) view.inputState.lastEscPress = Date.now()
+  if (event.keyCode == 27 && view.inputState.tabFocusMode != 0) view.inputState.tabFocusMode = Date.now() + 2000
   return false
 }
 
@@ -514,7 +539,11 @@ handlers.mousedown = (view, event: MouseEvent) => {
   if (style) {
     let mustFocus = !view.hasFocus
     view.inputState.startMouseSelection(new MouseSelection(view, event, style, mustFocus))
-    if (mustFocus) view.observer.ignore(() => focusPreventScroll(view.contentDOM))
+    if (mustFocus) view.observer.ignore(() => {
+      focusPreventScroll(view.contentDOM)
+      let active = view.root.activeElement
+      if (active && !active.contains(view.contentDOM)) (active as HTMLElement).blur()
+    })
     let mouseSel = view.inputState.mouseSelection
     if (mouseSel) {
       mouseSel.start(event)
@@ -537,8 +566,7 @@ function rangeForClick(view: EditorView, pos: number, bias: -1 | 1, type: number
   }
 }
 
-let insideY = (y: number, rect: Rect) => y >= rect.top && y <= rect.bottom
-let inside = (x: number, y: number, rect: Rect) => insideY(y, rect) && x >= rect.left && x <= rect.right
+let inside = (x: number, y: number, rect: Rect) => y >= rect.top && y <= rect.bottom && x >= rect.left && x <= rect.right
 
 // Try to determine, for the given coordinates, associated with the
 // given position, whether they are related to the element before or
@@ -557,8 +585,8 @@ function findPositionSide(view: EditorView, pos: number, x: number, y: number) {
   let after = line.coordsAt(off, 1)
   if (after && inside(x, y, after)) return 1
   // This is probably a line wrap point. Pick before if the point is
-  // beside it.
-  return before && insideY(y, before) ? -1 : 1
+  // above its bottom.
+  return before && before.bottom >= y ? -1 : 1
 }
 
 function queryPos(view: EditorView, event: MouseEvent): {pos: number, bias: 1 | -1} {
@@ -632,7 +660,8 @@ handlers.dragstart = (view, event: DragEvent) => {
   inputState.draggedContent = range
 
   if (event.dataTransfer) {
-    event.dataTransfer.setData("Text", view.state.sliceDoc(range.from, range.to))
+    event.dataTransfer.setData("Text", textFilter(view.state, clipboardOutputFilter,
+                                                  view.state.sliceDoc(range.from, range.to)))
     event.dataTransfer.effectAllowed = "copyMove"
   }
   return false
@@ -644,6 +673,7 @@ handlers.dragend = view => {
 }
 
 function dropText(view: EditorView, event: DragEvent, text: string, direct: boolean) {
+  text = textFilter(view.state, clipboardInputFilter, text)
   if (!text) return
   let dropPos = view.posAtCoords({x: event.clientX, y: event.clientY}, false)
 
@@ -743,7 +773,7 @@ function copiedRange(state: EditorState) {
     linewise = true
   }
 
-  return {text: content.join(state.lineBreak), ranges, linewise}
+  return {text: textFilter(state, clipboardOutputFilter, content.join(state.lineBreak)), ranges, linewise}
 }
 
 let lastLinewiseCopy: string | null = null
@@ -808,6 +838,7 @@ observers.blur = view => {
 }
 
 observers.compositionstart = observers.compositionupdate = view => {
+  if (view.observer.editContext) return // Composition handled by edit context
   if (view.inputState.compositionFirstChange == null)
     view.inputState.compositionFirstChange = true
   if (view.inputState.composing < 0) {
@@ -817,6 +848,7 @@ observers.compositionstart = observers.compositionupdate = view => {
 }
 
 observers.compositionend = view => {
+  if (view.observer.editContext) return // Composition handled by edit context
   view.inputState.composing = -1
   view.inputState.compositionEndedAt = Date.now()
   view.inputState.compositionPendingKey = true
@@ -843,7 +875,19 @@ observers.contextmenu = view => {
   view.inputState.lastContextMenu = Date.now()
 }
 
-handlers.beforeinput = (view, event) => {
+handlers.beforeinput = (view, event: InputEvent) => {
+  // In EditContext mode, we must handle insertReplacementText events
+  // directly, to make spell checking corrections work
+  if (event.inputType == "insertReplacementText" && view.observer.editContext) {
+    let text = event.dataTransfer?.getData("text/plain"), ranges = event.getTargetRanges()
+    if (text && ranges.length) {
+      let r = ranges[0]
+      let from = view.posAtDOM(r.startContainer, r.startOffset), to = view.posAtDOM(r.endContainer, r.endOffset)
+      applyDOMChangeInner(view, {from, to, insert: view.state.toText(text)}, null)
+      return true
+    }
+  }
+
   // Because Chrome Android doesn't fire useful key events, use
   // beforeinput to detect backspace (and possibly enter and delete,
   // but those usually don't even seem to fire beforeinput events at
