@@ -48,7 +48,8 @@ export const baseTheme = buildTheme("." + baseThemeID, {
     height: "100%",
     overflowX: "auto",
     position: "relative",
-    zIndex: 0
+    zIndex: 0,
+    overflowAnchor: "none",
   },
 
   ".cm-content": {
@@ -210,7 +211,8 @@ export const baseTheme = buildTheme("." + baseThemeID, {
     boxSizing: "border-box",
     position: "sticky",
     left: 0,
-    right: 0
+    right: 0,
+    zIndex: 300
   },
   "&light .cm-panels": {
     backgroundColor: "#f5f5f5",
