@@ -314,9 +314,10 @@ const tooltipPlugin = ViewPlugin.fromClass(class {
       let arrowHeight = arrow ? Arrow.Size : 0
       let width = size.right - size.left, height = knownHeight.get(tView) ?? size.bottom - size.top
       let offset = tView.offset || noOffset, ltr = this.view.textDirection == Direction.LTR
-      let left = size.width > space.right - space.left ? (ltr ? space.left : space.right - size.width)
-        : ltr ? Math.min(pos.left - (arrow ? Arrow.Offset : 0) + offset.x, space.right - width)
-        : Math.max(space.left, pos.left - width + (arrow ? Arrow.Offset : 0) - offset.x)
+      let left = size.width > space.right - space.left
+        ? (ltr ? space.left : space.right - size.width)
+        : ltr ? Math.max(space.left, Math.min(pos.left - (arrow ? Arrow.Offset : 0) + offset.x, space.right - width))
+        : Math.min(Math.max(space.left, pos.left - width + (arrow ? Arrow.Offset : 0) - offset.x), space.right - width)
       let above = this.above[i]
       if (!tooltip.strictSide && (above
             ? pos.top - (size.bottom - size.top) - offset.y < space.top
@@ -590,7 +591,9 @@ const showHoverTooltipHost = showTooltip.compute([showHoverTooltip], state => {
 
 const enum Hover { Time = 300, MaxDist = 6 }
 
-type HoverSource = (view: EditorView, pos: number, side: -1 | 1) => Tooltip | readonly Tooltip[] | null | Promise<Tooltip | readonly Tooltip[] | null>
+/// The type of function that can be used as a [hover tooltip
+/// source](#view.hoverTooltip^source).
+export type HoverTooltipSource = (view: EditorView, pos: number, side: -1 | 1) => Tooltip | readonly Tooltip[] | null | Promise<Tooltip | readonly Tooltip[] | null>
 
 class HoverPlugin {
   lastMove: {x: number, y: number, target: HTMLElement, time: number}
@@ -599,7 +602,7 @@ class HoverPlugin {
   pending: {pos: number} | null = null
 
   constructor(readonly view: EditorView,
-              readonly source: HoverSource,
+              readonly source: HoverTooltipSource,
               readonly field: StateField<readonly Tooltip[]>,
               readonly setHover: StateEffectType<readonly Tooltip[]>,
               readonly hoverTime: number) {
@@ -720,9 +723,14 @@ class HoverPlugin {
 const tooltipMargin = 4
 
 function isInTooltip(tooltip: HTMLElement, event: MouseEvent) {
-  let rect = tooltip.getBoundingClientRect()
-  return event.clientX >= rect.left - tooltipMargin && event.clientX <= rect.right + tooltipMargin &&
-    event.clientY >= rect.top - tooltipMargin && event.clientY <= rect.bottom + tooltipMargin
+  let {left, right, top, bottom} = tooltip.getBoundingClientRect(), arrow
+  if (arrow = tooltip.querySelector(".cm-tooltip-arrow")) {
+    let arrowRect = arrow.getBoundingClientRect()
+    top = Math.min(arrowRect.top, top)
+    bottom = Math.max(arrowRect.bottom, bottom)
+  }
+  return event.clientX >= left - tooltipMargin && event.clientX <= right + tooltipMargin &&
+    event.clientY >= top - tooltipMargin && event.clientY <= bottom + tooltipMargin
 }
 
 function isOverRange(view: EditorView, from: number, to: number, x: number, y: number, margin: number) {
@@ -744,8 +752,13 @@ function isOverRange(view: EditorView, from: number, to: number, x: number, y: n
 /// Note that all hover tooltips are hosted within a single tooltip
 /// container element. This allows multiple tooltips over the same
 /// range to be "merged" together without overlapping.
+///
+/// The return value is a valid [editor extension](#state.Extension)
+/// but also provides an `active` property holding a state field that
+/// can be used to read the currently active tooltips produced by this
+/// extension.
 export function hoverTooltip(
-  source: HoverSource,
+  source: HoverTooltipSource,
   options: {
     /// Controls whether a transaction hides the tooltip. The default
     /// is to not hide.
@@ -757,7 +770,7 @@ export function hoverTooltip(
     /// milliseconds. Defaults to 300ms.
     hoverTime?: number
   } = {}
-): Extension {
+): Extension & {active: StateField<readonly Tooltip[]>} {
   let setHover = StateEffect.define<readonly Tooltip[]>()
   let hoverState = StateField.define<readonly Tooltip[]>({
     create() { return [] },
@@ -790,11 +803,14 @@ export function hoverTooltip(
     provide: f => showHoverTooltip.from(f)
   })
 
-  return [
-    hoverState,
-    ViewPlugin.define(view => new HoverPlugin(view, source, hoverState, setHover, options.hoverTime || Hover.Time)),
-    showHoverTooltipHost
-  ]
+  return {
+    active: hoverState,
+    extension: [
+      hoverState,
+      ViewPlugin.define(view => new HoverPlugin(view, source, hoverState, setHover, options.hoverTime || Hover.Time)),
+      showHoverTooltipHost
+    ]
+  }
 }
 
 /// Get the active tooltip view for a given tooltip, if available.
