@@ -1,9 +1,10 @@
 import {Text, EditorState, ChangeSet, ChangeDesc, RangeSet, EditorSelection} from "@codemirror/state"
 import {Rect, isScrolledToBottom, getScale} from "./dom"
-import {HeightMap, HeightOracle, BlockInfo, MeasuredHeights, QueryType, heightRelevantDecoChanges} from "./heightmap"
+import {HeightMap, HeightOracle, BlockInfo, MeasuredHeights, QueryType, heightRelevantDecoChanges,
+        clearHeightChangeFlag, heightChangeFlag} from "./heightmap"
 import {decorations, ViewUpdate, UpdateFlag, ChangedRange, ScrollTarget, nativeSelectionHidden,
         contentAttributes} from "./extension"
-import {WidgetType, Decoration, DecorationSet} from "./decoration"
+import {WidgetType, Decoration, DecorationSet, BlockType} from "./decoration"
 import {EditorView} from "./editorview"
 import {Direction} from "./bidi"
 
@@ -22,7 +23,7 @@ function visiblePixelRange(dom: HTMLElement, paddingTop: number): Rect {
         left = Math.max(left, parentRect.left)
         right = Math.min(right, parentRect.right)
         top = Math.max(top, parentRect.top)
-        bottom = parent == dom.parentNode ? parentRect.bottom : Math.min(bottom, parentRect.bottom)
+        bottom = Math.min(parent == dom.parentNode ? win.innerHeight : bottom, parentRect.bottom)
       }
       parent = style.position == "absolute" || style.position == "fixed" ? elt.offsetParent : elt.parentNode
     } else if (parent.nodeType == 11) { // Shadow root
@@ -143,7 +144,7 @@ export class ViewState {
   defaultTextDirection: Direction = Direction.LTR
 
   // The main viewport for the visible part of the document
-  viewport: Viewport
+  viewport!: Viewport
   // If the main selection starts or ends outside of the main
   // viewport, extra single-line viewports are created for these
   // points, so that the DOM selection doesn't fall in a gap.
@@ -168,9 +169,11 @@ export class ViewState {
     this.stateDeco = state.facet(decorations).filter(d => typeof d != "function") as readonly DecorationSet[]
     this.heightMap = HeightMap.empty().applyChanges(this.stateDeco, Text.empty, this.heightOracle.setDoc(state.doc),
                                                     [new ChangedRange(0, 0, 0, state.doc.length)])
-    this.viewport = this.getViewport(0, null)
+    for (let i = 0; i < 2; i++) {
+      this.viewport = this.getViewport(0, null)
+      if (!this.updateForViewport()) break
+    }
     this.updateViewportLines()
-    this.updateForViewport()
     this.lineGaps = this.ensureLineGaps([])
     this.lineGapDeco = Decoration.set(this.lineGaps.map(gap => gap.draw(this, false)))
     this.computeVisibleRanges()
@@ -186,16 +189,21 @@ export class ViewState {
       }
     }
     this.viewports = viewports.sort((a, b) => a.from - b.from)
+    return this.updateScaler()
+  }
 
+  updateScaler() {
+    let scaler = this.scaler
     this.scaler = this.heightMap.height <= VP.MaxDOMHeight ? IdScaler :
       new BigScaler(this.heightOracle, this.heightMap, this.viewports)
+    return scaler.eq(this.scaler) ? 0 : UpdateFlag.Height
   }
 
   updateViewportLines() {
     this.viewportLines = []
     this.heightMap.forEachLine(this.viewport.from, this.viewport.to, this.heightOracle.setDoc(this.state.doc),
                                0, 0, block => {
-      this.viewportLines.push(this.scaler.scale == 1 ? block : scaleBlock(block, this.scaler))
+      this.viewportLines.push(scaleBlock(block, this.scaler))
     })
   }
 
@@ -209,9 +217,11 @@ export class ViewState {
       prevDeco, this.stateDeco, update ? update.changes : ChangeSet.empty(this.state.doc.length)))
     let prevHeight = this.heightMap.height
     let scrollAnchor = this.scrolledToBottom ? null : this.scrollAnchorAt(this.scrollTop)
+    clearHeightChangeFlag()
     this.heightMap = this.heightMap.applyChanges(this.stateDeco, update.startState.doc,
                                                  this.heightOracle.setDoc(this.state.doc), heightChanges)
-    if (this.heightMap.height != prevHeight) update.flags |= UpdateFlag.Height
+    if (this.heightMap.height != prevHeight || heightChangeFlag)
+      update.flags |= UpdateFlag.Height
     if (scrollAnchor) {
       this.scrollAnchorPos = update.changes.mapPos(scrollAnchor.from, -1)
       this.scrollAnchorHeight = scrollAnchor.top
@@ -224,11 +234,11 @@ export class ViewState {
     if (scrollTarget && (scrollTarget.range.head < viewport.from || scrollTarget.range.head > viewport.to) ||
         !this.viewportIsAppropriate(viewport))
       viewport = this.getViewport(0, scrollTarget)
-    let updateLines = !update.changes.empty || (update.flags & UpdateFlag.Height) ||
-      viewport.from != this.viewport.from || viewport.to != this.viewport.to
+    let viewportChange = viewport.from != this.viewport.from || viewport.to != this.viewport.to
     this.viewport = viewport
-    this.updateForViewport()
-    if (updateLines) this.updateViewportLines()
+    update.flags |= this.updateForViewport()
+    if (viewportChange || !update.changes.empty || (update.flags & UpdateFlag.Height)) this.updateViewportLines()
+
     if (this.lineGaps.length || this.viewport.to - this.viewport.from > (LG.Margin << 1))
       this.updateLineGaps(this.ensureLineGaps(this.mapLineGaps(this.lineGaps, update.changes)))
     update.flags |= this.computeVisibleRanges()
@@ -318,7 +328,7 @@ export class ViewState {
       if (dTop > 0 && dBottom > 0) bias = Math.max(dTop, dBottom)
       else if (dTop < 0 && dBottom < 0) bias = Math.min(dTop, dBottom)
 
-      oracle.heightChanged = false
+      clearHeightChangeFlag()
       for (let vp of this.viewports) {
         let heights = vp.from == this.viewport.from ? lineHeights : view.docView.measureVisibleLineHeights(vp)
         this.heightMap = (
@@ -326,15 +336,19 @@ export class ViewState {
                                                    [new ChangedRange(0, 0, 0, view.state.doc.length)]) : this.heightMap
         ).updateHeight(oracle, 0, refresh, new MeasuredHeights(vp.from, heights))
       }
-      if (oracle.heightChanged) result |= UpdateFlag.Height
+      if (heightChangeFlag) result |= UpdateFlag.Height
     }
 
     let viewportChange = !this.viewportIsAppropriate(this.viewport, bias) ||
       this.scrollTarget && (this.scrollTarget.range.head < this.viewport.from ||
                             this.scrollTarget.range.head > this.viewport.to)
-    if (viewportChange) this.viewport = this.getViewport(bias, this.scrollTarget)
-    this.updateForViewport()
-    if ((result & UpdateFlag.Height) || viewportChange) this.updateViewportLines()
+    if (viewportChange) {
+      if (result & UpdateFlag.Height) result |= this.updateScaler()
+      this.viewport = this.getViewport(bias, this.scrollTarget)
+      result |= this.updateForViewport()
+    }
+    if ((result & UpdateFlag.Height) || viewportChange)
+      this.updateViewportLines()
 
     if (this.lineGaps.length || this.viewport.to - this.viewport.from > (LG.Margin << 1))
       this.updateLineGaps(this.ensureLineGaps(refresh ? [] : this.lineGaps, view))
@@ -449,10 +463,10 @@ export class ViewState {
       gaps.push(gap)
     }
 
-    for (let line of this.viewportLines) {
-      if (line.length < doubleMargin) continue
+    let checkLine = (line: BlockInfo) => {
+      if (line.length < doubleMargin || line.type != BlockType.Text) return
       let structure = lineStructure(line.from, line.to, this.stateDeco)
-      if (structure.total < doubleMargin) continue
+      if (structure.total < doubleMargin) return
       let target = this.scrollTarget ? this.scrollTarget.range.head : null
       let viewFrom, viewTo
       if (wrapping) {
@@ -489,6 +503,11 @@ export class ViewState {
       if (viewFrom > line.from) addGap(line.from, viewFrom, line, structure)
       if (viewTo < line.to) addGap(viewTo, line.to, line, structure)
     }
+
+    for (let line of this.viewportLines) {
+      if (Array.isArray(line.type)) line.type.forEach(checkLine)
+      else checkLine(line)
+    }
     return gaps
   }
 
@@ -523,12 +542,16 @@ export class ViewState {
   }
 
   lineBlockAt(pos: number): BlockInfo {
-    return (pos >= this.viewport.from && pos <= this.viewport.to && this.viewportLines.find(b => b.from <= pos && b.to >= pos)) ||
+    return (pos >= this.viewport.from && pos <= this.viewport.to &&
+            this.viewportLines.find(b => b.from <= pos && b.to >= pos)) ||
       scaleBlock(this.heightMap.lineAt(pos, QueryType.ByPos, this.heightOracle, 0, 0), this.scaler)
   }
 
   lineBlockAtHeight(height: number): BlockInfo {
-    return scaleBlock(this.heightMap.lineAt(this.scaler.fromDOM(height), QueryType.ByHeight, this.heightOracle, 0, 0), this.scaler)
+    return (height >= this.viewportLines[0].top && height <= this.viewportLines[this.viewportLines.length - 1].bottom &&
+            this.viewportLines.find(l => l.top <= height && l.bottom >= height)) ||
+      scaleBlock(this.heightMap.lineAt(this.scaler.fromDOM(height), QueryType.ByHeight, this.heightOracle, 0, 0),
+                 this.scaler)
   }
 
   scrollAnchorAt(scrollTop: number) {
@@ -601,7 +624,8 @@ function find<T>(array: readonly T[], f: (value: T) => boolean): T | undefined {
 type YScaler = {
   toDOM(n: number): number
   fromDOM(n: number): number
-  scale: number
+  scale: number,
+  eq(other: YScaler): boolean
 }
 
 // Don't scale when the document height is within the range of what
@@ -609,7 +633,8 @@ type YScaler = {
 const IdScaler: YScaler = {
   toDOM(n: number) { return n },
   fromDOM(n: number) { return n },
-  scale: 1
+  scale: 1,
+  eq(other: YScaler) { return other == this }
 }
 
 // When the height is too big (> VP.MaxDOMHeight), scale down the
@@ -652,6 +677,12 @@ class BigScaler implements YScaler {
       base = vp.bottom; domBase = vp.domBottom
     }
   }
+
+  eq(other: YScaler) {
+    if (!(other instanceof BigScaler)) return false
+    return this.scale == other.scale && this.viewports.length == other.viewports.length &&
+      this.viewports.every((vp, i) => vp.from == other.viewports[i].from && vp.to == other.viewports[i].to)
+  }
 }
 
 function scaleBlock(block: BlockInfo, scaler: YScaler): BlockInfo {
